import json
from starlette.responses import Response as StarletteResponse
from starlette.background import BackgroundTask
from typing import Any, Literal, Mapping, get_args


class Responses[**T](StarletteResponse):

    def __init__(
        self,
        content: Any = None,
        status_code: int = 200,
        headers: Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(content, status_code, headers, media_type, background)


class PlainTextResponse[**T](Responses):
    media_type = "text/plain"

    def __init__(
        self,
        content: Any = None,
        status_code: int = 200,
        headers: Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(content, status_code, headers, media_type, background)

    def render(self, content: Any) -> bytes:
        if isinstance(content, bytes):
            return content
        if isinstance(content, str):
            return content.encode("utf-8")
        else:
            return super().render(content)


class HTMLResponse[**T](Responses):
    media_type = "text/html"

    def __init__(
        self,
        content: Any = None,
        status_code=Literal[200],
        headers: Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(content, 200, headers, media_type, background)

    def render(self, content: Any) -> bytes:
        if isinstance(content, bytes):
            return content
        if isinstance(content, str):
            return content.encode("utf-8")
        else:
            return super().render(content)


class JSONResponse[**T](Responses):
    media_type = "application/json"

    def __init__(
        self,
        content: Any = None,
        status_code: int = 200,
        headers: Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(content, status_code, headers, media_type, background)

    def render(self, content: Any) -> bytes:
        return json.dumps(content).encode("utf-8")


try:
    import orjson

    class ORJSONResponse[**T](Responses):

        def __init__(
            self,
            content: Any = None,
            status_code: int = 200,
            headers: Mapping[str, str] | None = None,
            media_type: str | None = None,
            background: BackgroundTask | None = None,
        ) -> None:
            super().__init__(content, status_code, headers, media_type, background)

        def render(self, content: Any) -> bytes:
            return orjson.dumps(content)

except ImportError as e:
    raise e



from typing import Any, Mapping
import pandas as pd
from pydantic import BaseModel
from starlette.background import BackgroundTask
from pypox.response.base import JSONResponse


class PandasResponse[**T](JSONResponse):

    def __init__(
        self,
        content: pd.DataFrame | Any = None,
        status_code: int = 200,
        headers: Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(content, status_code, headers, media_type, background)

    def render(self, content: pd.DataFrame) -> bytes:
        if issubclass(content.__class__, pd.DataFrame):
            return content.to_json(orient="records").encode("utf-8")
        return super().render(content)



from abc import ABC, abstractclassmethod
from pprint import pprint
from typing import Any, Optional
from starlette.routing import Route

try:
    from pydantic import BaseModel
except ImportError as e:
    raise e


class OpenAPI(BaseModel):
    """Represents an OpenAPI specification.

    Args:
        BaseModel (type): The base model class.

    Attributes:
        openapi (str): The OpenAPI version.
        info (Info): Information about the API.
        jsondialect (str): The JSON dialect used.
        servers (dict): The servers hosting the API.
        paths (dict): The API paths and operations.
        webhooks (dict): The API webhooks.
        components (dict): The API components.
        security (dict): The API security definitions.
        tag (dict): The API tags.
        externalDocs (dict): External documentation for the API.
    """

    openapi: str
    info: Optional["Info"] = None
    jsondialect: Optional[str] = None
    servers: Optional[list["Server"]] = None
    paths: Optional[dict[str, "PathItem"]] = None
    webhooks: Optional[dict[str, "PathItem | Reference"]] = None
    components: Optional["Components"] = None
    security: Optional["SecurityRequirement"] = None
    tag: Optional[list["Tag"]] = None
    externalDocs: Optional["ExternalDocumentation"] = None


class Info(BaseModel):
    """
    Represents information about the API.

    Args:
        BaseModel: The base model class.

    Attributes:
        title (str): The title of the API.
        description (str): The description of the API.
        termsOfService (str): The terms of service for the API.
        contact (Contact): The contact information for the API.
        license (License): The license information for the API.
        version (str): The version of the API.
    """

    title: Optional[str] = ""
    description: Optional[str] = ""
    termsOfService: Optional[str] = ""
    contact: Optional["Contact"] = None
    license: Optional["License"] = None
    version: Optional[str] = ""


class Contact(BaseModel):
    """
    Represents a contact.

    Args:
        name (str): The name of the contact.
        url (str): The URL of the contact.
        email (str): The email address of the contact.
    """

    name: Optional[str] = ""
    url: Optional[str] = ""
    email: Optional[str] = ""


class License(BaseModel):
    """Represents a license.

    Args:
        name (str): The name of the license.
        url (str): The URL of the license.
    """

    name: Optional[str] = ""
    url: Optional[str] = ""


class Server(BaseModel):
    """Represents a server.

    Args:
        BaseModel (type): The base model class.

    Attributes:
        url (str): The URL of the server.
        description (str): The description of the server.
        variables (dict): The server variables.
    """

    url: Optional[str] = None
    description: Optional[str] = None
    variables: Optional["ServerVariable"] = None


class ServerVariable(BaseModel):
    enum: Optional[list[str]] = None
    default: Optional[str] = None
    description: Optional[str] = None


class Components(BaseModel):
    schemas: Optional[dict[str, dict]] = None
    responses: Optional[dict[str, "Response | Reference"]] = None
    parameters: Optional[dict[str, "Parameter | Reference"]] = None
    examples: Optional[dict[str, dict]] = None
    requestBodies: Optional[dict[str, "RequestBody | Reference"]] = None
    securitySchemas: Optional[dict[str, dict]] = None
    links: Optional[dict[str, dict]] = None
    callbacks: Optional[dict[str, dict]] = None
    pathItems: Optional[dict[str, dict]] = None


class Response(BaseModel):
    description: Optional[str] = ""
    headers: Optional[dict[str, "Header | Reference"]] = None
    content: Optional[dict[str, "MediaType"]] = None
    links: Optional[dict[str, "Links | Reference"]] = None


class Parameter(BaseModel):
    name: Optional[str] = ""
    in_: Optional[str] = ""
    description: Optional[str] = ""
    required: Optional[bool] = None
    deprecated: Optional[bool] = None
    allowEmptyValue: Optional[bool] = None


class Header(Parameter):
    pass


class MediaType(BaseModel):
    schema_: Optional[dict] = None
    example: Optional[Any] = None
    examples: Optional[dict[str, "Example | Reference"]] = None
    encoding: Optional[dict[str, "Encoding"]] = None


class Encoding(BaseModel):
    contentType: Optional[str] = None
    headers: Optional[dict[str, "Header | Reference"]] = None
    style: Optional[str] = None
    explode: Optional[bool] = False
    allowReserved: Optional[bool] = False


class Example(BaseModel):
    summary: Optional[str] = ""
    description: Optional[str] = ""
    value: Optional[Any] = None
    externalValue: Optional[str] = ""


class RequestBody(BaseModel):
    description: Optional[str] = ""
    content: Optional[dict[str, "MediaType"]] = None
    required: Optional[bool] = False


class SecurityScheme(BaseModel):
    type: Optional[str] = ""
    description: Optional[str] = ""
    name: Optional[str] = ""
    in_: Optional[str] = ""
    _scheme: Optional[str] = ""
    bearerFormat: Optional[str] = ""
    flows: Optional[dict] = None
    openIdConnectUrl: Optional[str] = ""


class Links(BaseModel):
    operationRef: Optional[str] = None
    operationId: Optional[str] = None
    parameters: Optional[dict[str, Any]] = None
    requestBody: Optional["RequestBody"] = None
    description: Optional[str] = None
    server: Optional["Server"] = None


class Callbacks(BaseModel):
    pass


class PathItem(BaseModel):
    _ref: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    get: Optional["Operation"] = None
    post: Optional["Operation"] = None
    put: Optional["Operation"] = None
    delete: Optional["Operation"] = None
    options: Optional["Operation"] = None
    head: Optional["Operation"] = None
    patch: Optional["Operation"] = None
    trace: Optional["Operation"] = None
    servers: Optional[list["Server"]] = None
    parameters: Optional[list["Parameter"]] = None


class Operation(BaseModel):
    tag: Optional[list[str]] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    externalDocs: Optional["ExternalDocumentation"] = None
    operationId: Optional[str] = None
    parameters: Optional[list["Parameter | Reference"]] = None
    requestBody: Optional["RequestBody | Reference"] = None
    responses: Optional[dict[str, Response]] = None
    callbacks: Optional[dict[str, "Callbacks | Reference"]] = None
    deprecated: bool = False
    security: Optional[list["SecurityRequirement"]] = None
    servers: Optional[list[Server]] = None


class ExternalDocumentation(BaseModel):
    description: Optional[str] = None
    url: Optional[str] = None


class SecurityRequirement(BaseModel):
    pass


class Reference(BaseModel):
    _ref: Optional[str] = ""
    summary: Optional[str] = ""
    description: Optional[str] = ""


class Tag(BaseModel):
    name: Optional[str] = ""
    description: Optional[str] = ""
    externalDocs: Optional["ExternalDocumentation"] = None



"""
This module defines the conventions used in the pypox library for defining HTTP and websocket endpoints.

Classes:
    BaseConvention: Base class for conventions in the pypox library.
    HTTPConvention: Represents a convention for HTTP endpoints.
    WebsocketConvention: Represents a convention for websocket endpoints.
"""

from importlib.machinery import ModuleSpec
from inspect import iscoroutinefunction
import os
from types import ModuleType
from typing import Any, Callable, Generator
import importlib.util
from starlette.routing import BaseRoute, Route, WebSocketRoute, Router
from starlette.requests import Request
from starlette.responses import Response
from pypox.processor.base import (
    Processor,
)


class BaseConvention:
    """
    Base class for conventions in the pypox library.

    Attributes:
        name (str): The name of the convention.
        type (str): The type of the convention.
        files (list[str]): The list of files associated with the convention.
        callable (str): The name of the callable function in the module.
        directory (str): The directory path where the convention is defined.

    Methods:
        __init__(self,
        processor_func: list[BaseProcessor] | None,
        name: str, type: str,
        files: list[str],
        callable: str,
        directory: str) -> None:
            Initializes a BaseConvention object.
        __call__(self) -> list[BaseRoute]:
            Calls the convention and returns a list of BaseRoute objects.
        processor(self, func) -> Any:
            Decorator function that wraps the callable function with additional processing logic.
    """

    def __init__(
        self,
        name: str,
        _type: str,
        files: dict[str, str],
        _callable: str,
        directory: str,
    ) -> None:
        self._name = name
        self._type = _type
        self._files: dict[str, str] = files
        self._callable = _callable
        self._directory = directory
        self._processor_func: list = []

    def add_processor(self, processor_func: list | None) -> None:
        """
        Sets the processor function.

        Args:
            processor_func (list[BaseProcessor] | BaseProcessor | None): The list of processor functions.
        """
        if isinstance(processor_func, list):
            self._processor_func.extend(processor_func)
        elif isinstance(processor_func, Callable):
            self._processor_func.append(processor_func)
        else:
            raise ValueError("Processor function cannot be None")

    @property
    def processor_func(self) -> list:
        """
        The list of processor functions.
        """
        return self._processor_func

    @property
    def name(self) -> str:
        """
        The name of the convention.
        """
        return self._name

    @property
    def type(self) -> str:
        """
        The type of the convention.
        """
        return self._type

    @property
    def files(self) -> dict[str, str]:
        """
        The list of files associated with the convention.
        """
        return self._files

    @property
    def callable(self) -> str:
        """
        The name of the callable function in the module.
        """
        return self._callable

    @property
    def directory(self) -> str:
        """
        The directory path where the convention is defined.
        """
        return self._directory

    def __call__(self) -> Router:
        """
        Retrieves a list of BaseRoute objects based on the specified directory and files.

        Returns:
            A list of BaseRoute objects representing the routes defined in the specified directory and files.
        """
        router: list[BaseRoute] = []

        for root, file in self.walk():
            module_name: str = file.split(".")[0]
            module_path: str = os.path.join(root, file)

            module: ModuleType = self.load_module(module_name, module_path)

            route_path: str = self.create_route_path(self.directory, root)
            router.append(self.add_route(route_path, getattr(module, self.callable)))

        return Router(router)

    def add_route(
        self, route_path: str, func: Callable, methods: list[str] | None = None
    ) -> Route | WebSocketRoute:
        """Add a route to the application.

        This method is used to add a route to the application. It takes the route path,
        the handler function, and an optional list of HTTP methods as parameters.

        Args:
            route_path (str): # The path of the route.
            func (Callable): # The handler function for the route.
            methods (list[str] | None, optional): # The list of HTTP methods allowed for the route.
            If not provided, the default method is "GET". Defaults to None.
        Raises:
            ValueError: If the route type is invalid.

        Returns:
            Route | WebSocketRoute: The created route object.
        """
        if self.type == "http":
            if not methods:
                methods = ["GET"]
            return Route(
                route_path,
                endpoint=func,
                methods=methods,
            )
        if self.type == "websocket":
            return WebSocketRoute(
                route_path,
                endpoint=func,
            )
        raise ValueError("Invalid route type")

    def walk(self) -> Generator[tuple[str, str], Any, None]:
        """Recursively walks through the directory and yields tuples of root and file names.

        Yields:
            Generator[tuple[str, str], Any, None]: A generator that yields tuples of root and file names.
        """
        for root, _, files in os.walk(self.directory):
            for file in files:
                if file in self.files:
                    yield root, file

    def create_route_path(self, directory: str, root: str) -> str:
        """
        Create a route path by replacing the directory in the root path and performing some string replacements.

        Args:
            directory (str): The directory to be replaced in the root path.
            root (str): The root path.

        Returns:
            str: The modified route path.
        """
        return (
            root.replace(directory, "")
            .replace("\\", "/")
            .replace("[", "{")
            .replace("]", "}")
            + "/"
        )

    def load_module(self, module_name: str, module_path: str) -> ModuleType:
        """
        Load a module from a file location.

        Args:
            module_name (str): The name of the module.
            module_path (str): The path to the module file.

        Raises:
            ModuleNotFoundError: If the module is not found.
            AttributeError: If the callable is not found in the module.

        Returns:
            ModuleType: The loaded module.
        """
        spec: ModuleSpec | None = importlib.util.spec_from_file_location(
            module_name, module_path
        )
        if not spec:
            raise ModuleNotFoundError("Module not found")
        module: ModuleType = importlib.util.module_from_spec(spec)
        if not spec.loader:
            raise ModuleNotFoundError("Module not found")
        spec.loader.exec_module(module)
        if not hasattr(module, self.callable):
            raise AttributeError("Callable not found in module")
        return module


class HTTPConvetion(BaseConvention):
    """Represents a convention for HTTP endpoints.

    Args:
        BaseConvention (type): The base convention class.

    Attributes:
        name (str): The name of the convention.
        protocol (str): The protocol used by the convention.
        files (List[str]): The list of files associated with the convention.
        endpoint_type (str): The type of endpoint.
        directory (str): The directory where the convention is defined.
    """

    def __init__(
        self,
        directory: str,
    ) -> None:
        super().__init__(
            "HTTPConvention",
            "http",
            {
                "get.py": "GET",
                "post.py": "POST",
                "put.py": "PUT",
                "delete.py": "DELETE",
                "patch.py": "PATCH",
                "head.py": "HEAD",
                "options.py": "OPTIONS",
            },
            "endpoint",
            directory,
        )


class WebsocketConvention(BaseConvention):
    """
    A class representing a websocket convention.

    Args:
        BaseConvention (type): The base convention class.

    Attributes:
        directory (str): The directory where the convention is located.

    """

    def __init__(
        self,
        directory: str,
    ) -> None:
        """
        Initializes a new instance of the WebsocketConvention class.

        Args:
            directory (str): The directory where the convention is located.

        """
        super().__init__(
            "WebsocketConvention",
            "websocket",
            {"websocket.py": "WEBSOCKET"},
            "endpoint",
            directory,
        )


class HTMXConvention(BaseConvention):
    """
    A class representing the HTMX convention.

    Args:
        BaseConvention (type): The base convention class.
    """

    def __init__(self, directory: str) -> None:
        super().__init__(
            "HTMXConvention",
            "http",
            {"htmx.py": "GET"},
            "endpoint",
            directory,
        )



"""
    This module contains the base processor class for handling encoding and decoding of data.
"""

from ast import Param
from email.policy import default
from typing import (
    Iterable,
    NewType,
    Optional,
    Required,
    Self,
    TypeVar,
    Union,
    get_origin,
)
import asyncio
import inspect
from types import MappingProxyType, UnionType, coroutine
from typing import Any, Callable, Generic, Literal, Mapping, Sequence, get_args
from abc import ABC, abstractmethod
from pydantic import BaseModel
import starlette
from starlette.background import BackgroundTask
from starlette.requests import Request as StarletteRequest
from starlette.responses import (
    Response as StarletteResponse,
    JSONResponse,
    PlainTextResponse,
)
from starlette.exceptions import HTTPException
from starlette.websockets import WebSocket
from starlette.datastructures import Address
from starlette.routing import BaseRoute, Route, WebSocketRoute, Router
from orjson import dumps
from orjson import loads
from pypox.openapi.main import (
    Components,
    MediaType,
    Parameter as OASParameter,
    RequestBody as OASRequestBody,
    Response as OASResponse,
)

from functools import wraps
from starlette.applications import Starlette


class Processor:
    pass


class Encoder[T](Processor):

    def __init__(
        self,
        name: str,
        content: T = None,
    ) -> None:
        self.name: str = name
        self.content: T = content

    def __call__(self, request: StarletteRequest, _type: type) -> Self:

        return self

    def __oas__(self):
        return "OAS"

    def __bool__(self):
        return bool(self.content)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name}, content={self.content})"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: Any) -> bool:
        return self.content == other


class Response(Processor):
    pass


class RequestBody[T](Encoder):

    request_type: str = "body"
    media_type: str = "body"

    def __init__(self, name: str, content: Any) -> None:
        super().__init__(name, content)

    async def __call__(self, request: StarletteRequest, _type: type) -> Self:
        try:
            match self.request_type:
                case "body":
                    self.content = await request.body()
                case "application/x-www-form-urlencoded":
                    self.content = _type(**await request.form())
                case "application/json":
                    self.content = _type(**await request.json())
                case "multipart/form-data":
                    self.content = request.form()
                case "stream":
                    self.content = request.stream()
        except Exception as e:
            raise HTTPException(
                status_code=400, detail=f"Invalid request {self.__str__()}"
            )
        return self


class JSONBody[T](RequestBody):

    request_type = "application/json"
    media_type = "application/json"

    def __init__(self, name: str, content: T) -> None:
        super().__init__(
            name,
            content,
        )


class FormBody[T](RequestBody):

    request_type = "application/x-www-form-urlencoded"
    media_type = "application/x-www-form-urlencoded"

    def __init__(self, name: str, content: Any) -> None:
        super().__init__(name, content)


class MultiPartBody[T](RequestBody):

    request_type = "multipart/form-data"
    media_type = "multipart/form-data"

    def __init__(self, name: str, content: Any) -> None:
        super().__init__(name, content)


class Parameter[T](Encoder):

    request_type: str = ""

    def __init__(
        self,
        content: T = None,
        is_required: bool = True,
        name: str = "",
        description: str = "",
        deprecated: bool = False,
        allow_empty_value: bool = False,
    ) -> None:
        self.is_required = is_required
        self.description = description
        self.deprecated = deprecated
        self.allow_empty_value = allow_empty_value
        super().__init__(name=name, content=content)

    def __call__(self, request: StarletteRequest, _type: type) -> Self:
        data = getattr(request, self.request_type).get(
            self.name.lower().replace("_", "-")
        )
        if not data:
            if self.is_required:
                raise HTTPException(status_code=400, detail=f"{self.name} is required")
            return self
        self.content = _type(data)
        return self


class Query[T](Parameter):

    def __init__(
        self,
        content: T = None,
        is_required: bool = True,
        name: str = "",
        description: str = "",
        deprecated: bool = False,
        allow_empty_value: bool = False,
    ) -> None:
        self.request_type = "query_params"
        super().__init__(
            content=content,
            is_required=is_required,
            name=name,
            description=description,
            deprecated=deprecated,
            allow_empty_value=allow_empty_value,
        )


class Header[T](Parameter):

    def __init__(
        self,
        content: Any,
        is_required: bool = True,
        name: str = "",
        description: str = "",
        deprecated: bool = False,
        allow_empty_value: bool = False,
    ) -> None:
        self.request_type = "headers"
        super().__init__(
            content, is_required, name, description, deprecated, allow_empty_value
        )


class Cookie[T](Parameter):

    def __init__(
        self,
        content: Any,
        is_required: bool = True,
        name: str = "",
        description: str = "",
        deprecated: bool = False,
        allow_empty_value: bool = False,
    ) -> None:
        self.request_type = "cookies"
        super().__init__(
            content, is_required, name, description, deprecated, allow_empty_value
        )


class Path[T](Parameter):

    def __init__(
        self,
        content: Any,
        is_required: bool = True,
        name: str = "",
        description: str = "",
        deprecated: bool = False,
        allow_empty_value: bool = False,
    ) -> None:
        self.request_type = "path_params"
        super().__init__(
            content, is_required, name, description, deprecated, allow_empty_value
        )


class Client(Encoder):

    def __init__(self, name: str, content: Any) -> None:
        super().__init__(name, content)


def processor(func):

    @wraps(func)
    async def wrapper(request: StarletteRequest):

        func_params = {}

        for name, params in inspect.signature(func).parameters.items():

            if await is_optional(params):
                annotation = get_args(params.annotation)[0]
            else:
                annotation = params.annotation

            if annotation == StarletteRequest:
                func_params.update({name: request})
                continue

            if annotation == Client:
                func_params.update({name: Client(name, request.client)})
                continue

            if issubclass(get_origin(annotation), Parameter):
                func_params.update(
                    {name: await convert_paramaters(request, name, params)}
                )

            if issubclass(get_origin(annotation), RequestBody):
                func_params.update(
                    {name: await convert_request_body(request, name, params)}
                )
        if inspect.iscoroutinefunction(func):
            return await func(**func_params)
        return func(**func_params)

    return wrapper


async def convert_paramaters(
    request: StarletteRequest, name: str, parameter: inspect.Parameter
) -> Parameter:
    # TODO:
    # 1. check if paramater has a default value
    # 2. if it does, check if the default value is an instance of the paramater type
    # 3. if it is, return the default value
    # 4. if it is not, raise an error
    # check if the parameter has a default value
    parameter_type, sub_type, optional = await get_parameter_type(parameter)

    if optional:
        if not parameter.default == inspect._empty:
            data: Parameter[Any] = parameter_type(
                content=parameter.default.content,
                name=parameter.default.name or name,
                is_required=parameter.default.is_required,
            )
        else:
            data: Parameter[Any] = parameter_type(
                content=None, name=name, is_required=False
            )(request, sub_type)
    else:
        data = parameter_type(content=None, name=name)(request, sub_type)

    return data


async def convert_request_body(
    request: StarletteRequest, name: str, parameter: inspect.Parameter
) -> RequestBody:

    parameter_type, sub_type, optional = await get_request_body_type(parameter)

    if optional:
        if not parameter.default == inspect._empty:
            data: RequestBody[Any] = parameter_type(
                content=parameter.default.content,
                name=parameter.default.name or name,
            )
        else:
            data: RequestBody[Any] = parameter_type(content=None, name=name)
        return data
    else:
        data = parameter_type(content=None, name=name)

    return await data(request, sub_type)


async def get_parameter_type(
    parameter: inspect.Parameter,
) -> tuple[type[Parameter], type, bool]:
    if await is_optional(parameter):
        parameter_type: type[Parameter] = get_origin(get_args(parameter.annotation)[0])
        sub_type = get_args(get_args(parameter.annotation)[0])[0]
        optional = True
    else:
        parameter_type: type[Parameter] = get_origin(parameter.annotation)
        sub_type = get_args(parameter.annotation)[0]
        optional = False
    return parameter_type, sub_type, optional


async def get_request_body_type(
    parameter: inspect.Parameter,
) -> tuple[type[RequestBody], type, bool]:
    if await is_optional(parameter):
        parameter_type: type[RequestBody] = get_origin(
            get_args(parameter.annotation)[0]
        )
        sub_type = get_args(get_args(parameter.annotation)[0])[0]
        optional = True
    else:
        parameter_type: type[RequestBody] = get_origin(parameter.annotation)
        sub_type = get_args(parameter.annotation)[0]
        optional = False
    return parameter_type, sub_type, optional


async def is_optional(parameter: inspect.Parameter) -> bool:
    return get_origin(parameter.annotation) == Union and type(None) in get_args(
        parameter.annotation
    )



"""
    This module contains the processors used by the pypox framework.

    The processors handle different aspects of the HTTP request and response, such as cookies, headers, JSON data,
    path parameters, query parameters, and websockets.

    Optional dependencies include PydanticProcessor, HTMXProcessor, and JinjaProcessor.
"""
